from collections import defaultdict, OrderedDict

import numpy as np
import tensorflow as tf

from boml import utils, extension
from boml.load_data import BOMLExperiment

importlib = __import__("importlib")
boml_networks = importlib.import_module("boml.setup_model")
inner_grads = importlib.import_module("boml.lower_iter")
hyper_grads = importlib.import_module("boml.upper_iter")
boml_optimizer = importlib.import_module("boml.optimizer")


class BOMLOptimizer(object):
    """
    Wrapper for performing gradient-based metaparameter optimization
    """

    def __init__(
        self,
        method=None,
        inner_method=None,
        outer_method=None,
        truncate_iter=-1,
        experiments=[],
    ):
        """
        BMLHOptimizer is the main class in `pybml`, which takes responsibility for
        the whole process of model construnction and back propagation.

        :param Method: define basic method for following training process, it should be included in ['MetaInit', 'MetaRepr'],
         'MetaInit' type includes methods like 'MAML, FOMAML, TNet, WarpGrad'; 'MetaRepr' type includes methods like
         'BDA, RHG, Truncated, Implicit HG, DARTS';
        :param inner_method: method chosen for solving LLproblem, including ['Trad' ,'Simple', 'Aggr'], 'MetaRepr' type choose
        either 'Trad' for traditional optimization strategies or 'Aggr' for Gradient Aggragation optimization 'MetaInit' type
        should choose 'Simple', and set specific parameters for detailed method choices like FOMAML or TNet.
        :param outer_method: method chosen for solving LLproblem, including ['Reverse' ,'Simple', 'Forward', 'Implcit'],
        'MetaInit' type should choose Simple, and set specific parameters for detailed method choices like FOMAML
        :param truncate_iter: specific parameter for Truncated Reverse method, defining number of iterations to truncate
         in the Back propagation process
        :param experiments: list of experiment objects that has already been initialized
        :return :an initialized instance of BMLHOptimizer
        """
        assert method in ("MetaRepr", "MetaInit"), (
            "initialize method arguement, should be in list \an [MetaRepr,MetaInitl] "
            "MetaRepr based methods include [BDA,FHG,RHG,TRHG],"
            "HperOptim based methods include [MAML,FOMAML,MSGD]"
        )
        self._method = method
        assert inner_method in (
            "Aggr",
            "Simple",
            "Trad",
        ), "initialize method arguement, should be in list [Aggr, Simple, Trad]"
        self.inner_method = inner_method

        if self.inner_method == "Simple":
            assert outer_method == "Simple", (
                "Choose simple configuration of lower-level and upper-level "
                "calculation strategy for meta-initialization-based method"
            )
        elif self.inner_method == "Aggr":
            assert outer_method in ("Darts", "Reverse",), (
                "The bilevel aggregation strategy could choose Reverse Auto Differentiation"
                " or DARTS as upper-level calculation strategy"
            )
        else:
            assert outer_method in ("Reverse", "Implicit", "Darts",), (
                "Invalid combination of inner and upper strategies, "
                "please check initialization for different level of problems or "
                "extend the base classes to formulate your own problems definition"
            )
        self.outer_method = outer_method
        self._inner_gradient = getattr(
            inner_grads, "%s%s" % ("BOMLInnerGrad", self.inner_method)
        )
        if truncate_iter >= 0:
            self._outer_gradient = getattr(
                hyper_grads, "%s%s" % ("BOMLOuterGrad", self.outer_method)
            )(inner_method=self.inner_method, truncate_iter=truncate_iter)
        else:
            self._outer_gradient = getattr(
                hyper_grads, "%s%s" % ("BOMLOuterGrad", self.outer_method)
            )(inner_method=self.inner_method)
        self.io_opt = None
        self._learning_rate = None
        self._meta_learning_rate = None
        self.oo_opt = None
        self.data_set = None
        self._meta_model = None
        self._fin_hts = None
        self._global_step = None
        self._o_optim_dict = defaultdict(lambda: set())
        self._param_dict = OrderedDict()
        self._global_step = tf.get_variable(
            "global_step", initializer=0, trainable=False
        )
        self._inner_objective = None
        self._truncate_iterations = -1
        self.experiments = experiments
        # self.inner_objectives = []

    def meta_learner(
        self,
        _input,
        dataset,
        meta_model="v1",
        name="Hyper_Net",
        use_T=False,
        use_Warp=False,
        model_loss_func=utils.cross_entropy,
        outer_loss_func=utils.cross_entropy,
        **model_args
    ):
        """
        This method must be called once at first to build meta modules and initialize meta parameters and neural networks.
        :param _input: orginal input for neural network construction
        :param dataset: which dataset to use for training and testing. It should be initialized before being passed into the function
        :param meta_model: model chosen for neural network construction, 'v1' for C4L with fully connected layer,
        'v2' for Residual blocks with fully connected layer.
        :param name:  name for Meta model modules used for BMLNet initialization
        :param use_T: whether to use T layer for C4L neural networks
        :param model_loss_func: forms of loss functions of task-specific layer corresponding to 'WarpGrad' method, default to be cross-entropy
        :param outer_loss_func: forms of loss functions of task-specific layer corresponding to 'WarpGrad' method, default to be cross-entropy
        :return: BMLNet object containing the dict of hyper parameters
        """

        self.param_dict["use_T"] = use_T
        self.param_dict["use_Warp"] = use_Warp
        self.param_dict["model_loss_func"] = model_loss_func
        self.param_dict["outer_loss_func"] = outer_loss_func
        self.data_set = dataset
        assert meta_model.startswith(
            "V"
        ), "The dataset does not match the model chosen for meta_learner, V1,V2,...or Vk"
        meta_learner = getattr(
            boml_networks,
            "%s%s%s" % ("BOMLNet", dataset.train.name, self.method + meta_model),
        )(
            _input=_input,
            dim_output=dataset.train.dim_target,
            name=name,
            use_T=use_T,
            use_Warp=use_Warp,
            outer_method=self.outer_method,
            **model_args
        )
        self.param_dict["meta_model"] = meta_model
        self._meta_model = meta_learner

        return meta_learner

    def base_learner(
        self,
        _input,
        meta_learner,
        name="task_net",
        weights_initializer=tf.zeros_initializer,
    ):
        """
        This method has to be called for every experiment and takes responsibility
        for defining task-specific modules and inner optimizers.
        :param _input: orginal input for neural network construction of task-specific module
        :param meta_learner: returned value of Meta_model function, which is a instance of BMLNet or its child classes
        :param name: name for Base model modules used for BMLNet initialization
        :param weights_initializer: initializer function for task_specific network, called by 'MetaRepr' method
        :return: task-specific model part
        """
        if self.method == "MetaInit":
            base_learner = getattr(
                boml_networks,
                "%s%s%s"
                % (
                    "BOMLNet",
                    self.data_set.train.name,
                    self.method + self.param_dict["meta_model"],
                ),
            )(
                _input=_input,
                outer_param_dict=meta_learner.outer_param_dict,
                model_param_dict=meta_learner.model_param_dict,
                dim_output=meta_learner.dims[-1],
                name=name,
                use_T=meta_learner.use_T,
                outer_method=self.outer_method,
                use_Warp=meta_learner.use_Warp,
            )
        elif self.method == "MetaRepr":
            base_learner = getattr(boml_networks, "BOMLNetFeedForward")(
                _input=_input,
                dims=self.data_set.train.dim_target,
                output_weight_initializer=weights_initializer,
                name=name,
                use_T=meta_learner.use_T,
            )
        else:
            print(
                "initialize method arguement, should be in list \an [MetaRepr,MetaInitl]"
            )
            raise AssertionError
        return base_learner

    def ll_problem(
        self,
        inner_objective,
        learning_rate,
        T,
        inner_objective_optimizer="SGD",
        outer_objective=None,
        learn_lr=False,
        learn_st=False,
        alpha_init=0.0,
        s=1.0,
        t=1.0,
        learn_alpha=False,
        learn_alpha_itr=False,
        var_list=None,
        first_order=False,
        loss_func=utils.cross_entropy,
        momentum=0.5,
        experiment=None,
        **inner_kargs
    ):
        """
        After construction of neural networks, solutions to lower level problems should be regulated in LL_Problem.
        :param inner_objective: loss function for the inner optimization problem
        :param learning_rate: step size for inner loop optimization
        :param T: numbers of steps for inner gradient descent optimization
        :param inner_objective_optimizer: Optimizer type for the outer parameters, should be in list ['SGD','Momentum','Adam']
        :param outer_objective: loss function for the outer optimization problem, which need to be claimed in BDA agorithm
        :param learn_lr: BOOLEAN type, which determines whether to define learning rate as an outer parameter
        :param alpha_init: initial value of ratio of inner objective to outer objective in BDA algorithm
        :param learn_alpha: specify parameter for BDA algorithm to decide whether to initialize alpha as a hyper parameter
        :param learn_alpha_itr: parameter for BDA algorithm to specify whether to initialize alpha as a vector, of which
        every dimension's value is step-wise scale factor fot the optimization process
        :param learn_st: specify parameter for BDA algorithm to decide whether to initialize s and t as hyper parameters
        :param s: coefficients multiplied by outer objectives in BA algorithm, default to be 1.0
        :param t: coefficients multiplied by outer objectives in BA algorithm, default to be 1.0
        :param first_order: specific parameter to define whether to use implement first order MAML, default to be `FALSE`
        :param loss_func: specifying which type of loss function is used for the maml-based method, which should be
        consistent with the form to compute the inner objective
        :param momentum: specific parameter for Optimizer.BMLOptMomentum to set initial value of momentum
        :param experiment: instance of Experiment to use in the Lower Level Problem, especifially needed in the
         `MetaInit` type of methods.
        :param var_list: optional list of variables (of the inner optimization problem)from
        :param inner_kargs: optional arguments to pass to `py_bml.core.optimizer.minimize`
        :return: `BOMLInnerGrad` from boml.lower_iter.
        """
        if self._learning_rate is None:
            if learn_lr:
                self.param_dict["learn_lr"] = learn_lr
                self._learning_rate = tf.Variable(
                    name="lr",
                    initial_value=learning_rate,
                    dtype=tf.float32,
                    collections=extension.METAPARAMETERS_COLLECTIONS,
                    trainable=False,
                )
            else:
                self._learning_rate = tf.constant(learning_rate, name="lr")
            self.param_dict["learning_rate"] = self.learning_rate
        if self.io_opt is None:
            if self.inner_method == "Simple" or inner_objective_optimizer == "SGD":
                self.io_opt = getattr(
                    boml_optimizer, "%s%s" % ("BOMLOpt", inner_objective_optimizer)
                )(learning_rate=self._learning_rate, name=inner_objective_optimizer)
            elif inner_objective_optimizer == "Momentum":
                self.io_opt = getattr(boml_optimizer, "%s%s" % ("BOMLOpt", "Momentum"))(
                    learning_rate=self._learning_rate,
                    momentum=momentum,
                    name=inner_objective_optimizer,
                )
            else:
                self.io_opt = getattr(
                    boml_optimizer, "%s%s" % ("BOMLOpt", inner_objective_optimizer)
                )(learning_rate=self._learning_rate, name=inner_objective_optimizer)
        assert isinstance(self.io_opt, getattr(boml_optimizer, "BOMLOpt")), (
            "Must use an optimizer that extends "
            "the class boml.optimizers "
            "found {} instead".format(type(self.io_opt))
        )
        assert self._method in (
            "MetaRepr",
            "MetaInit",
        ), "illegal initialization value for argument:method, should be in [MetaRepr, MetaInit]"
        if self.method == "MetaRepr":
            if self.inner_method == "Aggr":
                assert (
                    outer_objective is not None
                ), "BDA must have upper-level loss functions passed to lower-level problems optimization process"

                if not (
                    ("s" in self._param_dict.keys()) or ("t" in self._param_dict.keys())
                ):
                    if learn_st:
                        s = extension.get_outerparameter("s", s)
                        t = extension.get_outerparameter("t", t)
                    else:
                        s = tf.constant(s, name="s")
                        t = tf.constant(t, name="t")
                    self._param_dict["s"] = s
                    self._param_dict["t"] = t

                if "alpha" not in self._param_dict.keys():
                    if learn_alpha_itr:
                        alpha_vec = np.ones((1, T), dtype=np.float32) * alpha_init
                        alpha = extension.get_outerparameter(
                            name="alpha", initializer=alpha_vec
                        )
                        t_tensor = tf.placeholder(
                            shape=(T, 1), dtype=tf.float32, name="t_tensor"
                        )
                    else:
                        alpha = (
                            extension.get_outerparameter(
                                initializer=alpha_init, name="alpha"
                            )
                            if learn_alpha
                            else tf.constant(alpha_init, name="alpha")
                        )
                        t_tensor = tf.placeholder(dtype=tf.float32, name="t_tensor")
                    self._param_dict["alpha"] = alpha
                    self._param_dict["t_tensor"] = t_tensor
        elif self.method == "MetaInit":
            self._param_dict["first_order"] = first_order

        assert isinstance(experiment, BOMLExperiment), (
            "MetaInit based methods require specialized "
            "task model for each generated task,"
            "please refer to basic instruction for modules of networks"
        )
        self._param_dict["loss_func"] = loss_func
        self._param_dict["experiment"] = experiment

        self._param_dict["T"] = T

        inner_grad = self._inner_gradient.compute_gradients(
            boml_opt=self.io_opt,
            loss_inner=inner_objective,
            loss_outer=outer_objective,
            param_dict=self._param_dict,
            var_list=var_list,
            **{
                inner_arg: inner_kargs[inner_arg]
                for inner_arg in set(inner_kargs.keys()) - set(self.param_dict.keys())
            }
        )
        if hasattr(inner_grad, "objective"):
            inner_objective = inner_grad.objective
            if self._inner_objective is None:
                self._inner_objective = [inner_objective]
            else:
                self._inner_objective = tf.concat(
                    (self._inner_objective, [inner_objective]), axis=0
                )
        else:
            pass

        return inner_grad

    def ul_problem(
        self,
        outer_objective,
        meta_learning_rate,
        inner_grad,
        mlr_decay=1.0e-5,
        meta_param=None,
        outer_objective_optimizer="Adam",
        epsilon=1.0,
        momentum=0.5,
        warp_lambda=1.0,
        tolerance=lambda _k: 0.1 * (0.9 ** _k),
    ):
        """
        Set the outer optimization problem and the descent procedure for the optimization of the
        outer parameters. Can be called at least once for every call of inner_problem, passing the resulting
         `OptimizerDict`. It can be called multiple times with different objective, optimizers and hyper_list s.

        :param outer_objective: scalar tensor for the outer objective
        :param meta_learning_rate: step size for outer loop optimization
        :param inner_grad: Returned value of py_bml.BMLHOptimizer.LLProblem()
        :param meta_param: optional list of outer parameters and model parameters
        :param outer_objective_optimizer: Optimizer type for the outer parameters,
        should be in list ['SGD','Momentum','Adam']
        :param epsilon: Float, cofffecients to be used in DARTS algorithm
        :param momentum: specific parameters to be used to initialize 'Momentum' algorithm
        :param tolerance: specific function template for Implicit HG Algorithm
        :param global_step: optional global step. By default tries to use the last variable
        in the collection GLOBAL_STEP
        :return: itself
        """
        if self._meta_learning_rate is None:
            self._meta_learning_rate = tf.train.inverse_time_decay(
                meta_learning_rate,
                self.global_step,
                decay_steps=1.0,
                decay_rate=mlr_decay,
            )
        if self.oo_opt is None:
            if outer_objective_optimizer == "Momentum":
                self.oo_opt = tf.train.MomentumOptimizer(
                    learning_rate=self._meta_learning_rate,
                    momentum=momentum,
                    name=outer_objective_optimizer,
                )
            elif outer_objective_optimizer == "Adam":
                self.oo_opt = tf.train.AdamOptimizer(
                    learning_rate=self._meta_learning_rate,
                    name=outer_objective_optimizer,
                )
            elif outer_objective_optimizer == "SGD":
                self.oo_opt = tf.train.GradientDescentOptimizer(
                    learning_rate=self._meta_learning_rate,
                    name=outer_objective_optimizer,
                )

        assert isinstance(
            self._outer_gradient, getattr(hyper_grads, "BOMLOuterGrad")
        ), "Wrong name for inner method,should be in list \n [Reverse, Simple, Forward, Implicit]"
        if self.param_dict["use_Warp"]:
            setattr(self.outergradient, "lambda", tf.cast(warp_lambda, tf.float32))

        if self.outer_method == "Darts" and (
            not hasattr(self.outergradient, "Epsilon")
        ):
            assert (
                self.param_dict["T"] == 1
            ), "Darts requires single gradient step to optimize task parameters"
            assert isinstance(
                self._outer_gradient, getattr(hyper_grads, "BOMLOuterGradDarts")
            ), "Wrong name for outer method,should be in list [Darts]"
            setattr(self.outergradient, "Epsilon", tf.cast(epsilon, tf.float32))
            setattr(self.outergradient, "param_dict", self.param_dict)
        if self.outer_method == "Implicit":
            self.outergradient.set_tolerance(tolerance=tolerance)
        meta_param = self.outergradient.compute_gradients(
            outer_objective,
            inner_grad,
            meta_param=meta_param,
            param_dict=self.param_dict,
        )
        self._o_optim_dict[self.oo_opt].update(meta_param)

        return self

    def aggregate_all(self, aggregation_fn=None, gradient_clip=None):
        """
        To be called when no more dynamics or problems will be added, computes the updates
        for the outer parameters. This behave nicely with global_variables_initializer.

        :param aggregation_fn: Optional operation to aggregate multiple hypergradients (for the same hyperparameter),
                                by (default: reduce_mean)
        :param gradient_clip: Optional operation like normalizing to be applied to hypergradients before performing
                            a descent step (default: nothing).

        :return: the run method of this object.
        """
        if self._fin_hts is None:
            _maybe_first_arg = lambda _v: _v[0] if isinstance(_v, tuple) else _v

            def maybe_first_arg(_v):
                return _v[0] if isinstance(_v, tuple) else _v

            self._fin_hts = tf.group(
                *[
                    maybe_first_arg(
                        opt.apply_gradients(
                            self.outergradient.hgrads_hvars(
                                meta_param=hll,
                                aggregation_fn=aggregation_fn,
                                gradient_clip=gradient_clip,
                            )
                        )
                    )
                    for opt, hll in self._o_optim_dict.items()
                ]
            )
            if self._global_step:
                with tf.control_dependencies([self._fin_hts]):
                    self._fin_hts = self._global_step.assign_add(1).op
        return self.run

    def run(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        session=None,
    ):
        """
        Run a whole iteration and updates the parameters of meta-learner.

        :param inner_objective_feed_dicts: feed dictionary for the Lower-Level problem. Can be a function of
                                            step, which accounts for, e.g. stochastic gradient descent.
        :param outer_objective_feed_dicts: feed dictionary for the Upper-Level optimization problem. Can be a function of
                                            hyper-iterations steps (i.e. global variable), which may account for, e.g.
                                            stochastic evaluation of upper objective.
        :param initializer_feed_dict:  an optional feed dictionary for the initialization of LL problem variables.
                                            Can be a function of hyper-iterations steps (i.e. global variable),
                                            which may account for, e.g. stochastic initialization.
        :param session: optional session
        """
        self._outer_gradient.apply_gradients(
            inner_objective_feed_dicts,
            outer_objective_feed_dicts,
            initializer_feed_dict,
            param_dict=self._param_dict,
            session=session,
            global_step=self._global_step,
        )
        ss = session or tf.get_default_session()

        def _opt_fd():
            _io_fd = (
                utils.maybe_call(
                    inner_objective_feed_dicts, utils.maybe_eval(self._global_step)
                )
                if inner_objective_feed_dicts
                else {}
            )
            _oo_fd = (
                utils.maybe_call(
                    outer_objective_feed_dicts, utils.maybe_eval(self._global_step)
                )
                if outer_objective_feed_dicts
                else {}
            )
            return utils.merge_dicts(_io_fd, _oo_fd)

        ss.run(self._hyperit, _opt_fd())


    @property
    def meta_model(self):
        """
        :return: the created BMLNet object
        """
        return self._meta_model

    @property
    def outergradient(self):
        """
        :return: the outergradient object underlying this wrapper.
        """
        return self._outer_gradient

    @property
    def innergradient(self):
        """
        :return: the innergradient object underlying this wrapper.
        """
        return self._inner_gradient

    @property
    def learning_rate(self):
        """
        :return: the outergradient object underlying this wrapper.
        """
        return self._learning_rate

    @property
    def meta_learning_rate(self):
        """
        :return: the outergradient object underlying this wrapper.
        """
        return self._meta_learning_rate

    @property
    def method(self):
        """
        :return: the method for whole algorithm.
        """
        return self._method

    @property
    def global_step(self):
        """
        :return: globalstep used in whole graph.
        """
        return self._global_step

    @property
    def param_dict(self):
        """
        :return: dict that holds hyper_params used in the inner optimization process.
        """
        return self._param_dict

    @property
    def _hyperit(self):
        """
        iteration of minimization of upperobjective(s), assuming the hyper-gradients are already computed.
        """
        assert (
            self._fin_hts is not None
        ), "Must call BMLHOptimizer.Aggregate_all before performing optimization."
        return self._fin_hts

    @property
    def inner_objectives(self):
        return self._outer_gradient.inner_objectives



"""
Module that contains datasets classes for managing data.
"""
import sys

import numpy as np
import scipy.sparse as sc_sp
import tensorflow as tf

try:
    import intervaltree as it
except ImportError:
    it = None

from boml.load_data.datasets.dl_utils import (
    maybe_cast_to_scalar,
    pad,
    stack_or_concat,
    vstack,
    convert_sparse_matrix_to_sparse_tensor,
    merge_dicts,
    get_rand_state,
    maybe_call,
)


class Datasets:
    """
    Simple object for standard datasets. Has the field `train` `validation` and `test` and support indexing
    """

    def __init__(self, train=None, validation=None, test=None):
        self.train = train
        self.validation = validation
        self.test = test
        self._lst = [train, validation, test]

    def setting(self):
        return {
            k: v.setting() if hasattr(v, "setting") else None
            for k, v in vars(self).items()
        }

    def __getitem__(self, item):
        return self._lst[item]

    def __len__(self):
        return len([_ for _ in self._lst if _ is not None])

    @property
    def name(self):
        return (
            self.train.name
        )  # could be that different datasets have different names....

    @staticmethod
    def from_list(list_of_datasets):
        """
        Generates a `Datasets` object from a list.

        :param list_of_datasets: list containing from one to three dataset
        :return:
        """
        train, valid, test = None, None, None
        train = list_of_datasets[0]
        if len(list_of_datasets) > 3:
            print("There are more then 3 Datasets here...")
            return list_of_datasets
        if len(list_of_datasets) > 1:
            test = list_of_datasets[-1]
            if len(list_of_datasets) == 3:
                valid = list_of_datasets[1]
        return Datasets(train, valid, test)

    @staticmethod
    def stack(*datasets_s):
        """
        Stack some datasets calling stack for each dataset.

        :param datasets_s:
        :return: a new dataset
        """
        return Datasets.from_list(
            [
                Dataset.stack(*[d[k] for d in datasets_s if d[k] is not None])
                for k in range(3)
            ]
        )


NAMED_SUPPLIER = {}


class Dataset:
    """
    Class for managing a single dataset, includes data and target fields and has some utility functions.
     It allows also to convert the dataset into tensors and to store additional information both on a
     per-example basis and general infos.
    """

    def __init__(
        self, data, target, sample_info=None, info=None, name=None, add_bias=False
    ):
        """

        :param data: Numpy array containing data
        :param target: Numpy array containing targets
        :param sample_info: either an array of dicts or a single dict, in which case it is cast to array of
                                  dicts.
        :param info: (optional) dictionary with further info about the dataset
        """
        self._tensor_mode = False
        self._data = data
        self._add_bias = add_bias
        if self._add_bias:
            assert isinstance(
                self.dim_data, int
            ), "Add bias not defined for non vector data"
            self._data = np.hstack((self.data, np.ones((self.num_examples, 1))))

        self._target = target
        if self._data is not None:  # in meta-dataset data and target can be unspecified
            if sample_info is None:
                sample_info = {}
            self.sample_info = (
                np.array([sample_info] * self.num_examples)
                if isinstance(sample_info, dict)
                else sample_info
            )

            assert self.num_examples == len(self.sample_info), (
                str(self.num_examples) + " " + str(len(self.sample_info))
            )
            assert self.num_examples == self._shape(self._target)[0]

        self.info = info or {}
        self.info.setdefault("_name", name)

    @property
    def name(self):
        return self.info["_name"]

    @property
    def bias(self):
        return self._add_bias

    def _shape(self, what):
        return what.get_shape().as_list() if self._tensor_mode else what.shape

    def setting(self):
        """
        for save setting purposes, does not save the actual data

        :return:
        """
        return {
            "num_examples": self.num_examples,
            "dim_data": self.dim_data,
            "dim_target": self.dim_target,
            "info": self.info,
        }

    @property
    def data(self):
        return self._data

    @property
    def target(self):
        return self._target

    @property
    def num_examples(self):
        """

        :return: Number of examples in this dataset
        """
        return self._shape(self.data)[0]

    @property
    def dim_data(self):
        """

        :return: The data dimensionality as an integer, if input are vectors, or a tuple in the general case
        """
        return maybe_cast_to_scalar(self._shape(self.data)[1:])

    @property
    def dim_target(self):
        """

        :return: The target dimensionality as an integer, if targets are vectors, or a tuple in the general case
        """
        shape = self._shape(self.target)
        return 1 if len(shape) == 1 else maybe_cast_to_scalar(shape[1:])

    def convert_to_tensor(self, keep_sparse=True):
        SPARSE_SCIPY_MATRICES = (sc_sp.csr.csr_matrix, sc_sp.coo.coo_matrix)
        matrices = ["_data", "_target"]
        for att in matrices:
            if keep_sparse and isinstance(
                self.__getattribute__(att), SPARSE_SCIPY_MATRICES
            ):
                self.__setattr__(
                    att,
                    convert_sparse_matrix_to_sparse_tensor(self.__getattribute__(att)),
                )
            else:
                self.__setattr__(
                    att,
                    tf.convert_to_tensor(self.__getattribute__(att), dtype=tf.float32),
                )
        self._tensor_mode = True

    def create_supplier(self, x, y, other_feeds=None, name=None):
        """
        Return a standard feed dictionary for this dataset.

        :param name: if not None, register this supplier in dict NAMED_SUPPLIERS (this can be useful for instance
                        when recording with rf.Saver)
        :param x: placeholder for data
        :param y: placeholder for target
        :param other_feeds: optional other feeds (dictionary or None)
        :return: a callable.
        """
        if not other_feeds:
            other_feeds = {}

        # noinspection PyUnusedLocal
        def _supplier(step=None):
            """

            :param step: unused, just for making it compatible with `HG` and `Saver`
            :return: the feed dictionary
            """
            if isinstance(self.data, WindowedData):
                data = self.data.generate_all()

            return {**{x: self.data, y: self.target}, **other_feeds}

        if name:
            NAMED_SUPPLIER[name] = _supplier

        return _supplier

    @staticmethod
    def stack(*datasets):
        """
        Assuming that the datasets have same structure, stacks data, targets and other info

        :param datasets:
        :return: stacked dataset
        """
        return Dataset(
            data=vstack([d.data for d in datasets]),
            target=stack_or_concat([d.target for d in datasets]),
            sample_info=np.concatenate([d.sample_info for d in datasets]),
            info={
                k: [d.info.get(k, None) for d in datasets]
                for k in merge_dicts(*[d.info for d in datasets])
            },
        )


class MetaDataset(Dataset):
    def __init__(self, info=None, name="MetaDataset", *args, **kwargs):
        super().__init__(None, None, None, info, name=name)
        self.args = args
        self.kwargs = kwargs

    def generate_datasets(self, rand=None, *args, **kwargs):
        """
        Generates and returns a single Datasets (possibly composed by training, validation and test sets)
        according to args and kwargs

        :param rand: random seed, state or None
        :param args:
        :param kwargs:
        :return: `Datasets`
        """
        raise NotImplementedError()

    def generate(self, count, batch_size=1, rand=None, *args, **kwargs):
        """
        Generator of datasets

        :param rand: random seed, state or None
        :param count: number of datasets to generate
        :param batch_size: number of episodes to generate at each call
        :param args:
        :param kwargs:
        :return: one or a list of Datasets objects
        """
        if not args:
            args = self.args
        if not kwargs:
            kwargs = self.kwargs
        rand = get_rand_state(rand)
        for _ in range(count):
            if batch_size == 1:
                yield self.generate_datasets(rand=rand, *args, **kwargs)
            else:
                yield self.generate_batch(batch_size, rand=rand, *args, **kwargs)

    def generate_batch(self, batch_size, rand=None, *args, **kwargs):
        """
        Generates a batch of Datasets
        """
        return [
            self.generate_datasets(rand, *args, **kwargs) for _ in range(batch_size)
        ]

    @property
    def dim_data(self, *args, **kwargs):
        return self.generate_datasets(*args, **kwargs).train.dim_data

    @property
    def dim_target(self, *args, **kwargs):
        return self.generate_datasets(*args, **kwargs).train.dim_target


class WindowedData(object):
    def __init__(self, data, row_sentence_bounds, window=5, process_all=False):
        """
        Class for managing windowed input data (like TIMIT).

        :param data: Numpy matrix. Each row should be an example data
        :param row_sentence_bounds:  Numpy matrix with bounds for padding. TODO add default NONE
        :param window: half-window size
        :param process_all: (default False) if True adds context to all data at object initialization.
                            Otherwise the windowed data is created in runtime.
        """
        assert it is not None, "NEED PACKAGE INTERVALTREE!"
        self.window = window
        self.data = data
        base_shape = self.data.shape
        self.shape = (base_shape[0], (2 * self.window + 1) * base_shape[1])
        self.tree = it.IntervalTree(
            [it.Interval(int(e[0]), int(e[1]) + 1) for e in row_sentence_bounds]
        )
        if process_all:
            print("adding context to all the dataset", end="- ")
            self.data = self.generate_all()
            print("DONE")
        self.process_all = process_all

    def generate_all(self):
        return self[:]

    def __getitem__(self, item):
        if hasattr(self, "process_all") and self.process_all:  # keep attr check!
            return self.data[item]
        if isinstance(item, int):
            return self.get_context(item=item)
        if isinstance(item, tuple):
            if len(item) == 2:
                rows, columns = item
                if isinstance(rows, int) and isinstance(columns, int):
                    # do you want the particular element?
                    return self.get_context(item=rows)[columns]
            else:
                raise TypeError("NOT IMPLEMENTED <|>")
            if isinstance(rows, slice):
                rows = range(*rows.indices(self.shape[0]))
            return np.vstack([self.get_context(r) for r in rows])[:, columns]
        else:
            if isinstance(item, slice):
                item = range(*item.indices(self.shape[0]))
            return np.vstack([self.get_context(r) for r in item])

    def __len__(self):
        return self.shape[0]

    def get_context(self, item):
        interval = list(self.tree[item])[0]
        # print(interval)
        left, right = interval[0], interval[1]
        left_pad = max(self.window + left - item, 0)
        right_pad = max(
            0, self.window - min(right, len(self) - 1) + item
        )  # this is to cope with reduce datasets
        # print(left, right, item)

        # print(left_pad, right_pad)
        base = np.concatenate(
            self.data[
                item - self.window + left_pad : item + self.window + 1 - right_pad
            ]
        )
        if left_pad:
            base = np.concatenate([pad(self.data[item], left_pad), base])
        if right_pad:
            base = np.concatenate([base, pad(self.data[item], right_pad)])
        return base


class ExampleVisiting:
    def __init__(self, dataset, batch_size, epochs=None, rnd=None):
        """
        Class for stochastic sampling of data points. It is most useful for feeding examples for the the
        training ops of `ReverseHG` or `ForwardHG`. Most notably, if the number of epochs is specified,
        the class takes track of the examples per mini-batches which is important for the backward pass
        of `ReverseHG` method.

        :param dataset: instance of `Dataset` class
        :param batch_size:
        :param epochs: number of epochs (can be None, in which case examples are
                        fed continuously)
        """
        self.dataset = dataset
        self.batch_size = batch_size
        self.epochs = epochs
        self.T = int(np.ceil(dataset.num_examples / batch_size))
        if self.epochs:
            self.T *= self.epochs

        self.rnd = get_rand_state(rnd)

        self.training_schedule = None
        self.iter_per_epoch = int(dataset.num_examples / batch_size)

    def setting(self):
        excluded = ["training_schedule", "datasets"]
        dictionary = {k: v for k, v in vars(self).items() if k not in excluded}
        if hasattr(self.dataset, "setting"):
            dictionary["dataset"] = self.dataset.setting()
        return dictionary

    def generate_visiting_scheme(self):
        """
        Generates and stores example visiting scheme, as a numpy array of integers.

        :return: self
        """

        def all_indices_shuffled():
            _res = list(range(self.dataset.num_examples))
            self.rnd.shuffle(_res)
            return _res

        # noinspection PyUnusedLocal
        _tmp_ts = np.concatenate(
            [all_indices_shuffled() for _ in range(self.epochs or 1)]
        )
        self.training_schedule = (
            _tmp_ts
            if self.training_schedule is None
            else np.concatenate([self.training_schedule, _tmp_ts])
        )  # do not discard previous schedule,
        # this should allow backward passes of arbitrary length

        return self

    def create_supplier(self, x, y, *other_feeds, name=None):
        return self.create_feed_dict_supplier(x, y, *other_feeds, name=name)

    def create_feed_dict_supplier(self, x, y, *other_feeds, name=None):
        """

        :param name: optional name for this supplier
        :param x: placeholder for independent variable
        :param y: placeholder for dependent variable
        :param other_feeds: dictionary of other feeds (e.g. dropout factor, ...) to add to the input output
                            feed_dict
        :return: a function that generates a feed_dict with the right signature for Reverse and Forward HyperGradient
                    classes
        """

        def _training_supplier(step=None):
            nonlocal other_feeds

            if step >= self.T:
                if step % self.T == 0:
                    if self.epochs:
                        print(
                            "WARNING: End of the training scheme reached."
                            "Generating another scheme.",
                            file=sys.stderr,
                        )
                    self.generate_visiting_scheme()
                step %= self.T

            if self.training_schedule is None:
                # print('visiting scheme not yet generated!')
                self.generate_visiting_scheme()

            # noinspection PyTypeChecker
            nb = self.training_schedule[
                step
                * self.batch_size : min(
                    (step + 1) * self.batch_size, len(self.training_schedule)
                )
            ]

            bx = self.dataset.data[nb, :]
            by = self.dataset.target[nb, :]

            # if lambda_feeds:  # this was previous implementation... dunno for what it was used for
            #     lambda_processed_feeds = {k: v(nb) for k, v in lambda_feeds.items()}  previous implementation...
            #  looks like lambda was
            # else:
            #     lambda_processed_feeds = {}
            return merge_dicts(
                {x: bx, y: by}, *[maybe_call(of, step) for of in other_feeds]
            )

        if name:
            NAMED_SUPPLIER[name] = _training_supplier

        return _training_supplier



from collections import OrderedDict

import tensorflow as tf

from boml import utils
from boml.upper_iter.outer_grad import BOMLOuterGrad


class BOMLOuterGradSimple(BOMLOuterGrad):
    def __init__(self, inner_method="Simple", history=None, name="BMLOuterOptSimple"):
        super(BOMLOuterGradSimple, self).__init__(name)
        self._inner_method = inner_method
        self._history = history or []
        self._reverse_initializer = tf.no_op()
        self.warp_lambda = tf.cast(1.0, tf.float32)
        self.reptile_initializer = tf.no_op()

    # noinspection SpellCheckingInspection

    def compute_gradients(
        self, outer_objective, inner_grad, meta_param=None, param_dict=OrderedDict()
    ):
        """
        Function that adds to the computational graph all the operations needend for computing
        the hypergradients in a "dynamic" way, without unrolling the entire optimization graph.
        The resulting computation, while being roughly 2x more expensive then unrolling the
        optimizaiton dynamics, requires much less (GPU) memory and is more flexible, allowing
        to set a termination condition to the parameters optimizaiton routine.

        :param inner_grad: OptimzerDict object resulting from the inner objective optimization.
        :param outer_objective: A loss function for the outer parameters (scalar tensor)
        :param meta_param: Optional list of outer parameters to consider. If not provided will get all variables in the
                            hyperparameter collection in the current scope.
        :param param_dict: dictionary of parameters to specify different methods
        :return: list of outer parameters involved in the computation
        """
        meta_param = super(BOMLOuterGradSimple, self).compute_gradients(
            outer_objective, inner_grad, meta_param
        )

        with tf.variable_scope(outer_objective.op.name):
            """
            if len(meta_param) == len(list(optimizer_dict.state)):
                doo_dhypers = tf.gradients(outer_objective,list(optimizer_dict.state))
                #doo_dhypers = tf.gradients(outer_objective, list(optimizer_dict.model_fast_weights.values()))
            else:
                """
            if param_dict["use_Warp"]:
                doo_dhypers = [
                    self.warp_lambda * outer_param
                    for outer_param in inner_grad.outer_param_tensor
                ] + inner_grad.model_param_tensor
                doo_dhypers += tf.gradients(
                    outer_objective, meta_param[len(doo_dhypers) :]
                )
            else:
                doo_dhypers = tf.gradients(
                    outer_objective,
                    list(inner_grad.state) + meta_param[len(inner_grad.state) :],
                )

            for h, doo_dh in zip(meta_param, doo_dhypers):
                assert doo_dh is not None, BOMLOuterGrad._ERROR_HYPER_DETACHED.format(
                    doo_dh
                )
                self._hypergrad_dictionary[h].append(doo_dh)

            return meta_param

    def apply_gradients(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        param_dict=OrderedDict(),
        global_step=None,
        session=None,
    ):

        ss = session or tf.get_default_session()

        self._history.clear()

        _fd = utils.maybe_call(initializer_feed_dict, utils.maybe_eval(global_step, ss))
        self._save_history(ss.run(self.initialization, feed_dict=_fd))

    def _save_history(self, weights):
        self._history.append(weights)



from setuptools import setup,find_packages
from os import path

current_directory = path.abspath(path.dirname(__file__))

with open("README.rst", "r") as fh:
    long_description = fh.read()

def get_install_requirements():
    requirements_path = path.join(current_directory, "requirements.txt")
    with open(requirements_path, encoding="utf-8") as fp:
        return fp.read().splitlines()


setup(
    name="boml",
    version="0.1.2a1",
    packages=find_packages(),
    long_description=long_description,
    url="https://github.com/dut-media-lab/BOML",
    license="MIT",
    keywords=[
    "meta-learning",
    "bilevel-optimization",
    "few-shot-lerning",
    "python",
    "Deep learning"],
    install_requires=get_install_requirements(),
    classifiers=[
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    ],
    python_requires='>=3.5,<3.8',
    author="Yaohua Liu, Risheng Liu",
    author_email="liuyaohua@mail.dlut.edu.cn",
    description="A Bilevel Optimizer Library in Python for Meta Learning",
)



from collections import OrderedDict

import tensorflow as tf
from boml.extension import GraphKeys
from boml.setup_model.network import BOMLNet
from boml.utils import as_tuple_or_list, remove_from_collection


class BOMLNetFeedForward(BOMLNet):
    def __init__(
        self,
        _input,
        dims,
        task_parameter=OrderedDict(),
        name="BMLNetFeedForward",
        activation=tf.nn.relu,
        var_collections=tf.GraphKeys.MODEL_VARIABLES,
        output_weight_initializer=tf.contrib.layers.xavier_initializer(tf.float32),
        data_type=tf.float32,
        reuse=False,
    ):
        """

        :param _input: original input for the FeedForward network,
        which is also output of the convolutional neural networks
        :param dims: the dimension of the final output
        :param task_parameter: dictionary to store the task-specific parameters
        :param name: name for the task-specific network
        :param activation: activation function, default to be tf.nn.relu
        :param var_collections: collections to manage the parameters of
         base-learner in the computational graph
        :param output_weight_initializer: function to initialize the weights of FeedForward network
        :param data_type: default to be tf.float32
        :param reuse: whether to reuse the created parameters in the named scope
        :param use_T:
        """
        self.dims = as_tuple_or_list(dims)
        self.activation = activation
        self.data_type = data_type
        self.task_parameter = task_parameter
        self.var_collections = var_collections
        self.output_weight_initializer = output_weight_initializer
        super().__init__(
            _input=_input, name=name, var_collections=var_collections, reuse=reuse,
        )
        if not reuse:
            print(name, "MODEL CREATED")

    def _forward(self):

        if len(self.task_parameter) == 0:
            self.create_initial_parameter()
        self + tf.add(
            tf.matmul(self.out, self.task_parameter["fc_weight"]),
            self.task_parameter["fc_bias"],
        )

    def create_initial_parameter(self):
        self.task_parameter = OrderedDict()
        self.task_parameter["fc_weight"] = tf.get_variable(
            "fc_weight",
            shape=[self.layers[-1].shape.as_list()[-1], self.dims[-1]],
            initializer=self.output_weight_initializer,
            dtype=self.data_type,
        )
        self.task_parameter["fc_bias"] = tf.get_variable(
            "fc_bias",
            [self.dims[-1]],
            initializer=tf.zeros_initializer,
            dtype=self.data_type,
        )
        [
            tf.add_to_collections(self.var_collections, initial_param)
            for initial_param in self.task_parameter.values()
        ]
        remove_from_collection(
            GraphKeys.GLOBAL_VARIABLES, *self.task_parameter.values()
        )

    def re_forward(self, new_input=None, task_parameter=OrderedDict()):
        return BOMLNetFeedForward(
            new_input if new_input is not None else self.layers[0],
            dims=self.dims,
            task_parameter=self.task_parameter
            if len(task_parameter) == 0
            else task_parameter,
            name=self.name,
            activation=self.activation,
            data_type=self.data_type,
            var_collections=self.var_collections,
            output_weight_initializer=self.output_weight_initializer,
            reuse=True,
        )



from boml.load_data.experiment import BOMLExperiment
from boml.load_data import (
    meta_mini_imagenet,
    meta_omniglot,
    OmniglotMetaDataset,
    ImageNetMetaDataset,
    mnist,
)
from boml.setup_model.meta_init_v1 import (
    BOMLNetMetaInitV1,
    BOMLNetOmniglotMetaInitV1,
    BOMLNetMiniMetaInitV1,
)
from boml.setup_model.meta_repr_v1 import (
    BOMLNetMetaReprV1,
    BOMLNetMiniMetaReprV1,
    BOMLNetOmniglotMetaReprV1,
)
from boml.setup_model.meta_init_v2 import (
    BOMLNetOmniglotMetaInitV2,
    BOMLNetMiniMetaInitV2,
)
from boml.setup_model.meta_repr_v2 import (
    BOMLNetOmniglotMetaReprV2,
    BOMLNetMiniMetaReprV2,
)
from boml.setup_model.network import BOMLNet
from boml.setup_model.feedforward import BOMLNetFeedForward
from boml.optimizer import BOMLOptSGD, BOMLOptMomentum, BOMLOptAdam
from boml.boml_optimizer import BOMLOptimizer
from boml.extension import get_outerparameter
from boml import utils
from boml import extension



from collections import OrderedDict

import numpy as np
import tensorflow as tf
from tensorflow.contrib.layers.python import layers

from boml.extension import GraphKeys
from boml.setup_model import network_utils
from boml.setup_model.network import BOMLNet
from boml.setup_model.network_utils import as_tuple_or_list
from boml.utils import remove_from_collection


class BOMLNetMetaInitV1(BOMLNet):
    def __init__(
        self,
        _input,
        dim_output,
        name="BOMLNetMetaInitV1",
        outer_param_dict=OrderedDict(),
        model_param_dict=None,
        task_parameter=None,
        use_T=False,
        use_Warp=False,
        outer_method="Simple",
        activation=tf.nn.relu,
        var_collections=tf.GraphKeys.MODEL_VARIABLES,
        conv_initializer=tf.contrib.layers.xavier_initializer_conv2d(tf.float32),
        output_weight_initializer=tf.contrib.layers.xavier_initializer(tf.float32),
        norm=layers.batch_norm,
        data_type=tf.float32,
        channels=1,
        dim_hidden=[64, 64, 64, 64],
        kernel=3,
        max_pool=False,
        reuse=False,
    ):
        self.task_parameter = task_parameter
        self.kernel = kernel
        self.channels = channels
        self.dims = as_tuple_or_list(dim_output)
        self.dim_hidden = dim_hidden
        self.datatype = data_type
        self.batch_norm = norm
        self.max_pool = max_pool
        self.stride = [1, 2, 2, 1]
        self.no_stride = [1, 1, 1, 1]
        self.activation = activation
        self.bias_initializer = tf.zeros_initializer(tf.float32)
        self.conv_initializer = conv_initializer
        self.output_weight_initializer = output_weight_initializer
        self.outer_method = outer_method
        self.use_T = use_T
        self.use_Warp = use_Warp

        super().__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            var_collections=var_collections,
            name=name,
            model_param_dict=model_param_dict,
            reuse=reuse,
        )

        # variables from batch normalization
        self.betas = self.filter_vars("beta")
        # moving mean and variance (these variables should be used at inference time... so must save them)
        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")

        if not reuse:  # these calls might print a warning... it's not a problem..
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.betas)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_means)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_variances)
            print(name, "MODEL CREATED")

    def create_outer_parameters(self, var_collections=GraphKeys.METAPARAMETERS):
        for i in range(len(self.dim_hidden)):
            self.outer_param_dict["conv" + str(i)] = network_utils.get_conv_weight(
                self, i=i, initializer=self.conv_initializer
            )
            self.outer_param_dict["bias" + str(i)] = network_utils.get_bias_weight(
                self, i=i, initializer=self.bias_initializer
            )
        if self.max_pool:
            self.outer_param_dict["w" + str(len(self.dim_hidden))] = tf.get_variable(
                "w" + str(len(self.dim_hidden)),
                [self.dim_hidden[-1] * 5 * 5, self.dims[-1]],
                initializer=self.output_weight_initializer,
            )
            self.outer_param_dict["bias" + str(len(self.dim_hidden))] = tf.get_variable(
                "bias" + str(len(self.dim_hidden)),
                [self.dims[-1]],
                initializer=self.bias_initializer,
                dtype=self.datatype,
            )
        else:
            self.outer_param_dict["w" + str(len(self.dim_hidden))] = tf.get_variable(
                "w" + str(len(self.dim_hidden)),
                [self.dim_hidden[-1], self.dims[-1]],
                initializer=tf.random_normal_initializer,
            )
            self.outer_param_dict["bias" + str(len(self.dim_hidden))] = tf.get_variable(
                "bias" + str(len(self.dim_hidden)),
                [self.dims[-1]],
                initializer=self.bias_initializer,
                dtype=self.datatype,
            )
        [
            tf.add_to_collections(var_collections, hyper)
            for hyper in self.outer_param_dict.values()
        ]

        if len(self.model_param_dict) == 0 and callable(
            getattr(self, "create_model_parameters", None)
        ):
            self.create_model_parameters()

        return self.outer_param_dict

    def create_model_parameters(self, var_collections=GraphKeys.METAPARAMETERS):
        if self.use_T:
            # hyper parameters of transformation layer
            for i in range(len(self.dim_hidden)):
                self.model_param_dict[
                    "conv" + str(i) + "_z"
                ] = network_utils.get_identity(
                    self.dim_hidden[0], name="conv" + str(i) + "_z", conv=True
                )
            self.model_param_dict[
                "w" + str(len(self.dim_hidden)) + "_z"
            ] = network_utils.get_identity(
                self.dims[-1], name="w" + str(len(self.dim_hidden)) + "_z", conv=False
            )
        elif self.use_Warp:
            for i in range(len(self.dim_hidden)):
                self.model_param_dict[
                    "conv" + str(i) + "_z"
                ] = network_utils.get_warp_weight(
                    self, i, self.conv_initializer
                )
                self.model_param_dict[
                    "bias" + str(i) + "_z"
                ] = network_utils.get_warp_bias(
                    self, i, self.bias_initializer
                )
        [
            tf.add_to_collections(var_collections, model_param)
            for model_param in self.model_param_dict.values()
        ]
        return self.model_param_dict

    def _forward(self):
        if self.task_parameter is None:
            self.task_parameter = self.create_initial_parameter(
                primary_outerparameter=self.outer_param_dict
            )

        for i in range(len(self.dim_hidden)):
            if self.use_T:
                self + network_utils.conv_block_t(
                    self,
                    self.task_parameter["conv" + str(i)],
                    self.task_parameter["bias" + str(i)],
                    self.model_param_dict["conv" + str(i) + "_z"],
                )
            elif self.use_Warp:
                self + network_utils.conv_block_warp(
                    self,
                    self.task_parameter["conv" + str(i)],
                    self.task_parameter["bias" + str(i)],
                    self.model_param_dict["conv" + str(i) + "_z"],
                    self.model_param_dict["bias" + str(i) + "_z"],
                )
            else:
                self + network_utils.conv_block(
                    self,
                    self.task_parameter["conv" + str(i)],
                    self.task_parameter["bias" + str(i)],
                )

        if self.max_pool:
            self + tf.reshape(
                self.out, [-1, np.prod([int(dim) for dim in self.out.get_shape()[1:]])]
            )
            self + tf.add(
                tf.matmul(
                    self.out, self.task_parameter["w" + str(len(self.dim_hidden))]
                ),
                self.task_parameter["bias" + str(len(self.dim_hidden))],
            )
        else:
            self + tf.add(
                tf.matmul(
                    tf.reduce_mean(self.out, [1, 2]),
                    self.task_parameter["w" + str(len(self.dim_hidden))],
                ),
                self.task_parameter["bias" + str(len(self.dim_hidden))],
            )

        if self.use_T:
            self + tf.matmul(
                self.out, self.model_param_dict["w" + str(len(self.dim_hidden)) + "_z"]
            )

    def re_forward(self, new_input=None, task_parameter=OrderedDict()):
        return BOMLNetMetaInitV1(
            _input=new_input if new_input is not None else self.layers[0],
            dim_output=self.dims[-1],
            name=self.name,
            activation=self.activation,
            outer_param_dict=self.outer_param_dict,
            model_param_dict=self.model_param_dict,
            task_parameter=self.task_parameter
            if len(task_parameter.keys()) == 0
            else task_parameter,
            use_T=self.use_T,
            use_Warp=self.use_Warp,
            outer_method=self.outer_method,
            var_collections=self.var_collections,
            dim_hidden=self.dim_hidden,
            output_weight_initializer=self.output_weight_initializer,
            max_pool=self.max_pool,
            reuse=True,
        )


def BOMLNetOmniglotMetaInitV1(
    _input,
    dim_output,
    outer_param_dict=OrderedDict(),
    model_param_dict=OrderedDict(),
    batch_norm=layers.batch_norm,
    name="BOMLNetOmniglotMetaInitV1",
    outer_method="Simple",
    use_T=False,
    use_Warp=False,
    **model_args
):
    return BOMLNetMetaInitV1(
        _input=_input,
        name=name,
        dim_output=dim_output,
        model_param_dict=model_param_dict,
        outer_method=outer_method,
        outer_param_dict=outer_param_dict,
        norm=batch_norm,
        use_T=use_T,
        use_Warp=use_Warp,
        **model_args
    )


def BOMLNetMiniMetaInitV1(
    _input,
    dim_output,
    outer_param_dict=OrderedDict(),
    model_param_dict=OrderedDict(),
    batch_norm=layers.batch_norm,
    name="BOMLNetMetaInitV1",
    outer_method="Simple",
    use_T=False,
    use_Warp=False,
    **model_args
):
    return BOMLNetMetaInitV1(
        _input=_input,
        name=name,
        dim_output=dim_output,
        use_T=use_T,
        use_Warp=use_Warp,
        outer_param_dict=outer_param_dict,
        model_param_dict=model_param_dict,
        outer_method=outer_method,
        norm=batch_norm,
        channels=3,
        dim_hidden=[32, 32, 32, 32],
        max_pool=True,
        **model_args
    )



from boml.upper_iter.outer_grad import BOMLOuterGrad
from boml.upper_iter.reverse import BOMLOuterGradReverse
from boml.upper_iter.simple import BOMLOuterGradSimple
from boml.upper_iter.implicit import BOMLOuterGradImplicit
from boml.upper_iter.darts import BOMLOuterGradDarts



from __future__ import print_function, absolute_import, division

from collections import OrderedDict

import tensorflow as tf

# import numpy as np
GRADIENT_NONE_MESSAGE = (
    "WARNING: the gradient w.r.t.the tf.Variable\n {}\n is None;\n "
    "Check the computational graph of the inner objective, and be sure you\n"
    "are not considering including variables that should not be there among the\n"
    "inner variables."
)


class BOMLInnerGradTrad(object):
    def __init__(self, update_op, dynamics, objective):
        """
        :param update_op: the tf operation to perform updates
        :param dynamics: the iterative formats of dynamical system
        :param objective: Lower-Level objective
        """
        self._updates_op = update_op
        self._dynamics = dynamics
        self._iteration = None
        self._initialization = None
        self._init_dyn = (
            None  # for phi_0 (will be a dictionary (state-variable, phi_0 op)
        )
        self.objective = objective
        self.bml_opt = None

    @staticmethod
    def compute_gradients(
        boml_opt,
        loss_inner,
        loss_outer=None,
        param_dict=OrderedDict(),
        var_list=None,
        **inner_kargs
    ):
        """
        :param boml_opt: instance of modified optimizers in the `optimizer` module
        :param loss_inner: Lower-Level objectives
        :param loss_outer: Upper-Level objectives
        :param param_dict: dictionary of general parameters for different algorithms
        :param var_list: the list of parameters in the base-learner
        :param inner_kargs: optional arguments for tensorflow optimizers, like global_step, gate_gradients
        :return: initialized instance of inner_grad for UL optimization
        """

        minimize_kargs = {
            inner_arg: inner_kargs[inner_arg]
            for inner_arg in set(inner_kargs.keys()) - set(param_dict.keys())
        }

        assert loss_inner is not None, "argument:inner_objective must be initialized"
        update_op, dynamics = boml_opt.minimize(
            loss_inner=loss_inner, var_list=var_list, *minimize_kargs
        )

        return BOMLInnerGradTrad(
            update_op=update_op, dynamics=dynamics, objective=loss_inner
        )  # loss_inner

    @property
    def apply_updates(self):
        """
        Descent step, as returned by `tf.train.Optimizer.apply_gradients`.
        :return:
        """
        return self._updates_op

    @property
    def iteration(self):
        """
        Performs a descent step (as return by `tf.train.Optimizer.apply_gradients`) and computes the values of
        the variables after it.

        :return: A list of operation that, after performing one iteration, return the value of the state variables
                    being optimized (possibly including auxiliary variables)
        """
        if self._iteration is None:
            with tf.control_dependencies([self._updates_op]):  # ?
                self._iteration = (
                    self._state_read()
                )  # performs an iteration and returns the
                # value of all variables in the state (ordered according to dyn)

        return self._iteration

    @property
    def initialization(self):
        """
        :return: a list of operations that return the values of the state variables for this
                    learning dynamics after the execution of the initialization operation. If
                    an initial dynamics is set, then it also executed.
        """
        if self._initialization is None:
            with tf.control_dependencies([tf.variables_initializer(self.state)]):
                if (
                    self._init_dyn is not None
                ):  # create assign operation for initialization
                    self._initialization = [
                        k.assign(v) for k, v in self._init_dyn.items()
                    ]
                    # return these new initialized values (and ignore variable initializers)
                else:
                    self._initialization = (
                        self._state_read()
                    )  # initialize state variables and
                    # return the initialized value

        return self._initialization

    @property
    def dynamics(self):
        """
        :return: A generator for the dynamics (state_variable_{k+1})
        """
        return self._dynamics.values()

    @property
    def state(self):
        """
        :return: A generator for all the state variables (optimized variables and possibly auxiliary variables)
        being optimized
        """
        return self._dynamics.keys()  # overridden in Adam

    def _state_read(self):
        """
        :return: generator of read value op for the state variables
        """
        return [
            v.read_value() for v in self.state
        ]  # not sure about read_value vs value

    def state_feed_dict(self, his):
        """
        Builds a feed dictionary of (past) states
        """
        return {v: his[k] for k, v in enumerate(self.state)}

    @property
    def init_dynamics(self):
        """
        :return: The initialization dynamics if it has been set, or `None` otherwise.
        """
        return None if self._init_dyn is None else list(self._init_dyn.items())

    def __lt__(self, other):  # make OptimizerDict sortable
        # TODO be sure that this is consistent
        assert isinstance(other, BOMLInnerGradTrad)
        return hash(self) < hash(other)

    def __len__(self):
        return len(self._dynamics)



"""
Simple container for useful quantities for a supervised learning experiment, where data is managed
with feed dictionary
"""
import tensorflow as tf


class BOMLExperiment:
    def __init__(self, datasets, dtype=tf.float32):
        """
        Define data formats and manage attributes of the generated tasks.
        :param datasets: instance of Meta Dataset
        :param dtype: default to be tf.float32
        """

        self.datasets = datasets
        self.x = tf.placeholder(dtype, name="x", shape=self._compute_input_shape())
        self.y = tf.placeholder(dtype, name="y", shape=self._compute_output_shape())
        self.x_ = tf.placeholder(dtype, name="x_", shape=self._compute_input_shape())
        self.y_ = tf.placeholder(dtype, name="y_", shape=self._compute_output_shape())
        self.dtype = dtype
        self.model = None
        self.errors = {}
        self.scores = {}
        self.optimizers = {}

    # noinspection PyBroadException
    def _compute_input_shape(self):
        """
        :return: compute the shape of input according to the data formats
        """
        sh = self.datasets.train.dim_data
        return (None, sh) if isinstance(sh, int) else (None,) + sh

    # noinspection PyBroadException
    def _compute_output_shape(self):
        """
        :return: compute the shape of output according to the data formats
        """
        sh = self.datasets.train.dim_target
        return (None, sh) if isinstance(sh, int) else (None,) + sh
